/* The copyright in this software is beinOMg made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.cpp
    \brief    prediction class
*/

#include <memory.h>
#include "TComPrediction.h"
#include "TComTU.h"

//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Tables
// ====================================================================================================================

const UChar TComPrediction::m_aucIntraFilter[MAX_NUM_CHANNEL_TYPE][MAX_INTRA_FILTER_DEPTHS] =
{
  { // Luma
    10, //4x4
    7, //8x8
    1, //16x16
    0, //32x32
    10, //64x64
  },
  { // Chroma
    10, //4xn
    7, //8xn
    1, //16xn
    0, //32xn
    10, //64xn
  }

};

// ====================================================================================================================
// Constructor / destructor / initialize
// ====================================================================================================================

TComPrediction::TComPrediction()
: m_pLumaRecBuffer(0)
, m_iLumaRecStride(0)
{
  for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
  {
    for(UInt buf=0; buf<2; buf++)
    {
      m_piYuvExt[ch][buf] = NULL;
    }
  }
#if SCM__R0348_PALETTE_MODE
  m_iPLTErrLimit = 3;
#endif
}

TComPrediction::~TComPrediction()
{
  destroy();
}

Void TComPrediction::destroy()
{
  for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
  {
    for(UInt buf=0; buf<NUM_PRED_BUF; buf++)
    {
      delete [] m_piYuvExt[ch][buf];
      m_piYuvExt[ch][buf] = NULL;
    }
  }

  for(UInt i=0; i<NUM_REF_PIC_LIST_01; i++)
  {
    m_acYuvPred[i].destroy();
  }

  m_cYuvPredTemp.destroy();

  if( m_pLumaRecBuffer )
  {
    delete [] m_pLumaRecBuffer;
    m_pLumaRecBuffer = 0;
  }
  m_iLumaRecStride = 0;

  for (UInt i = 0; i < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; i++)
  {
    for (UInt j = 0; j < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; j++)
    {
      m_filteredBlock[i][j].destroy();
    }
    m_filteredBlockTmp[i].destroy();
  }
}

Void TComPrediction::initTempBuff(ChromaFormat chromaFormatIDC)
{
  // if it has been initialised before, but the chroma format has changed, release the memory and start again.
  if( m_piYuvExt[COMPONENT_Y][PRED_BUF_UNFILTERED] != NULL && m_cYuvPredTemp.getChromaFormat()!=chromaFormatIDC)
  {
    destroy();
  }

  if( m_piYuvExt[COMPONENT_Y][PRED_BUF_UNFILTERED] == NULL ) // check if first is null (in which case, nothing initialised yet)
  {
    Int extWidth  = MAX_CU_SIZE + 16;
    Int extHeight = MAX_CU_SIZE + 1;

    for (UInt i = 0; i < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; i++)
    {
      m_filteredBlockTmp[i].create(extWidth, extHeight + 7, chromaFormatIDC);
      for (UInt j = 0; j < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; j++)
      {
        m_filteredBlock[i][j].create(extWidth, extHeight, chromaFormatIDC);
      }
    }

    m_iYuvExtSize = (MAX_CU_SIZE*2+1) * (MAX_CU_SIZE*2+1);
    for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
    {
      for(UInt buf=0; buf<NUM_PRED_BUF; buf++)
      {
        m_piYuvExt[ch][buf] = new Pel[ m_iYuvExtSize ];
      }
    }

    // new structure
    for(UInt i=0; i<NUM_REF_PIC_LIST_01; i++)
    {
      m_acYuvPred[i] .create( MAX_CU_SIZE, MAX_CU_SIZE, chromaFormatIDC );
    }

    m_cYuvPredTemp.create( MAX_CU_SIZE, MAX_CU_SIZE, chromaFormatIDC );
  }


  if (m_iLumaRecStride != (MAX_CU_SIZE>>1) + 1)
  {
    m_iLumaRecStride =  (MAX_CU_SIZE>>1) + 1;
    if (!m_pLumaRecBuffer)
    {
      m_pLumaRecBuffer = new Pel[ m_iLumaRecStride * m_iLumaRecStride ];
    }
  }
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

// Function for calculating DC value of the reference samples used in Intra prediction
//NOTE: Bit-Limit - 25-bit source
Pel TComPrediction::predIntraGetPredValDC( const Pel* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, ChannelType channelType, ChromaFormat format, Bool bAbove, Bool bLeft )
{
  assert(iWidth > 0 && iHeight > 0);
  Int iInd, iSum = 0;
  Pel pDcVal;

  if (bAbove)
  {
    for (iInd = 0;iInd < iWidth;iInd++)
    {
      iSum += pSrc[iInd-iSrcStride];
    }
  }
  if (bLeft)
  {
    for (iInd = 0;iInd < iHeight;iInd++)
    {
      iSum += pSrc[iInd*iSrcStride-1];
    }
  }

  if (bAbove && bLeft)
  {
    pDcVal = (iSum + iWidth) / (iWidth + iHeight);
  }
  else if (bAbove)
  {
    pDcVal = (iSum + iWidth/2) / iWidth;
  }
  else if (bLeft)
  {
    pDcVal = (iSum + iHeight/2) / iHeight;
  }
  else
  {
    pDcVal = pSrc[-1]; // Default DC value already calculated and placed in the prediction array if no neighbors are available
  }

  return pDcVal;
}

// Function for deriving the angular Intra predictions

/** Function for deriving the simplified angular intra predictions.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 * \param dirMode the intra prediction mode index
 * \param blkAboveAvailable boolean indication if the block above is available
 * \param blkLeftAvailable boolean indication if the block to the left is available
 *
 * This function derives the prediction samples for the angular mode based on the prediction direction indicated by
 * the prediction mode index. The prediction direction is given by the displacement of the bottom row of the block and
 * the reference row above the block in the case of vertical prediction or displacement of the rightmost column
 * of the block and reference column left from the block in the case of the horizontal prediction. The displacement
 * is signalled at 1/32 pixel accuracy. When projection of the predicted pixel falls inbetween reference samples,
 * the predicted value for the pixel is linearly interpolated from the reference samples. All reference samples are taken
 * from the extended main reference.
 */
//NOTE: Bit-Limit - 25-bit source
Void TComPrediction::xPredIntraAng(       Int bitDepth,
                                    const Pel* pSrc,     Int srcStride,
                                          Pel* pTrueDst, Int dstStrideTrue,
                                          UInt uiWidth, UInt uiHeight, ChannelType channelType, ChromaFormat format,
                                          UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable
                                  , const Bool bEnableEdgeFilters
                                  )
{
  Int width=Int(uiWidth);
  Int height=Int(uiHeight);

  // Map the mode index to main prediction direction and angle
  assert( dirMode != PLANAR_IDX ); //no planar
  const Bool modeDC        = dirMode==DC_IDX;

  // Do the DC prediction
  if (modeDC)
  {
    const Pel dcval = predIntraGetPredValDC(pSrc, srcStride, width, height, channelType, format, blkAboveAvailable, blkLeftAvailable);

    for (Int y=height;y>0;y--, pTrueDst+=dstStrideTrue)
    {
      for (Int x=0; x<width;) // width is always a multiple of 4.
      {
        pTrueDst[x++] = dcval;
      }
    }
  }
  else // Do angular predictions
  {
    const Bool       bIsModeVer         = (dirMode >= 18);
    const Int        intraPredAngleMode = (bIsModeVer) ? (Int)dirMode - VER_IDX :  -((Int)dirMode - HOR_IDX);
    const Int        absAngMode         = abs(intraPredAngleMode);
    const Int        signAng            = intraPredAngleMode < 0 ? -1 : 1;
    const Bool       edgeFilter         = bEnableEdgeFilters && isLuma(channelType) && (width <= MAXIMUM_INTRA_FILTERED_WIDTH) && (height <= MAXIMUM_INTRA_FILTERED_HEIGHT);

    // Set bitshifts and scale the angle parameter to block size
    static const Int angTable[9]    = {0,    2,    5,   9,  13,  17,  21,  26,  32};
    static const Int invAngTable[9] = {0, 4096, 1638, 910, 630, 482, 390, 315, 256}; // (256 * 32) / Angle
    Int invAngle                    = invAngTable[absAngMode];
    Int absAng                      = angTable[absAngMode];
    Int intraPredAngle              = signAng * absAng;

    Pel* refMain;
    Pel* refSide;

    Pel  refAbove[2*MAX_CU_SIZE+1];
    Pel  refLeft[2*MAX_CU_SIZE+1];

    // Initialise the Main and Left reference array.
    if (intraPredAngle < 0)
    {
      const Int refMainOffsetPreScale = (bIsModeVer ? height : width ) - 1;
      const Int refMainOffset         = height - 1;
      for (Int x=0;x<width+1;x++)
      {
        refAbove[x+refMainOffset] = pSrc[x-srcStride-1];
      }
      for (Int y=0;y<height+1;y++)
      {
        refLeft[y+refMainOffset] = pSrc[(y-1)*srcStride-1];
      }
      refMain = (bIsModeVer ? refAbove : refLeft)  + refMainOffset;
      refSide = (bIsModeVer ? refLeft  : refAbove) + refMainOffset;

      // Extend the Main reference to the left.
      Int invAngleSum    = 128;       // rounding for (shift by 8)
      for (Int k=-1; k>(refMainOffsetPreScale+1)*intraPredAngle>>5; k--)
      {
        invAngleSum += invAngle;
        refMain[k] = refSide[invAngleSum>>8];
      }
    }
    else
    {
      for (Int x=0;x<2*width+1;x++)
      {
        refAbove[x] = pSrc[x-srcStride-1];
      }
      for (Int y=0;y<2*height+1;y++)
      {
        refLeft[y] = pSrc[(y-1)*srcStride-1];
      }
      refMain = bIsModeVer ? refAbove : refLeft ;
      refSide = bIsModeVer ? refLeft  : refAbove;
    }

    // swap width/height if we are doing a horizontal mode:
    Pel tempArray[MAX_CU_SIZE*MAX_CU_SIZE];
    const Int dstStride = bIsModeVer ? dstStrideTrue : MAX_CU_SIZE;
    Pel *pDst = bIsModeVer ? pTrueDst : tempArray;
    if (!bIsModeVer)
    {
      std::swap(width, height);
    }

    if (intraPredAngle == 0)  // pure vertical or pure horizontal
    {
      for (Int y=0;y<height;y++)
      {
        for (Int x=0;x<width;x++)
        {
          pDst[y*dstStride+x] = refMain[x+1];
        }
      }

      if (edgeFilter)
      {
        for (Int y=0;y<height;y++)
        {
          pDst[y*dstStride] = Clip3 (0, ((1 << bitDepth) - 1), pDst[y*dstStride] + (( refSide[y+1] - refSide[0] ) >> 1) );
        }
      }
    }
    else
    {
      Pel *pDsty=pDst;

      for (Int y=0, deltaPos=intraPredAngle; y<height; y++, deltaPos+=intraPredAngle, pDsty+=dstStride)
      {
        const Int deltaInt   = deltaPos >> 5;
        const Int deltaFract = deltaPos & (32 - 1);

        if (deltaFract)
        {
          // Do linear filtering
          const Pel *pRM=refMain+deltaInt+1;
          Int lastRefMainPel=*pRM++;
          for (Int x=0;x<width;pRM++,x++)
          {
            Int thisRefMainPel=*pRM;
            pDsty[x+0] = (Pel) ( ((32-deltaFract)*lastRefMainPel + deltaFract*thisRefMainPel +16) >> 5 );
            lastRefMainPel=thisRefMainPel;
          }
        }
        else
        {
          // Just copy the integer samples
          for (Int x=0;x<width; x++)
          {
            pDsty[x] = refMain[x+deltaInt+1];
          }
        }
      }
    }

    // Flip the block if this is the horizontal mode
    if (!bIsModeVer)
    {
      for (Int y=0; y<height; y++)
      {
        for (Int x=0; x<width; x++)
        {
          pTrueDst[x*dstStrideTrue] = pDst[x];
        }
        pTrueDst++;
        pDst+=dstStride;
      }
    }
  }
}

Void TComPrediction::predIntraAng( const ComponentID compID, UInt uiDirMode, Pel* piOrg /* Will be null for decoding */, UInt uiOrgStride, Pel* piPred, UInt uiStride, TComTU &rTu, Bool bAbove, Bool bLeft, const Bool bUseFilteredPredSamples, const Bool bUseLosslessDPCM )
{
  const ChromaFormat   format      = rTu.GetChromaFormat();
  const ChannelType    channelType = toChannelType(compID);
  const TComRectangle &rect        = rTu.getRect(isLuma(compID) ? COMPONENT_Y : COMPONENT_Cb);
  const Int            iWidth      = rect.width;
  const Int            iHeight     = rect.height;

  assert( g_aucConvertToBit[ iWidth ] >= 0 ); //   4x  4
  assert( g_aucConvertToBit[ iWidth ] <= 5 ); // 128x128
  //assert( iWidth == iHeight  );

        Pel *pDst = piPred;

  // get starting pixel in block
  const Int sw = (2 * iWidth + 1);

  if ( bUseLosslessDPCM )
  {
    const Pel *ptrSrc = getPredictorPtr( compID, false );
    // Sample Adaptive intra-Prediction (SAP)
    if (uiDirMode==HOR_IDX)
    {
      // left column filled with reference samples
      // remaining columns filled with piOrg data (if available).
      for(Int y=0; y<iHeight; y++)
      {
        piPred[y*uiStride+0] = ptrSrc[(y+1)*sw];
      }
      if (piOrg!=0)
      {
        piPred+=1; // miss off first column
        for(Int y=0; y<iHeight; y++, piPred+=uiStride, piOrg+=uiOrgStride)
        {
          memcpy(piPred, piOrg, (iWidth-1)*sizeof(Pel));
        }
      }
    }
    else // VER_IDX
    {
      // top row filled with reference samples
      // remaining rows filled with piOrd data (if available)
      for(Int x=0; x<iWidth; x++)
      {
        piPred[x] = ptrSrc[x+1];
      }
      if (piOrg!=0)
      {
        piPred+=uiStride; // miss off the first row
        for(Int y=1; y<iHeight; y++, piPred+=uiStride, piOrg+=uiOrgStride)
        {
          memcpy(piPred, piOrg, iWidth*sizeof(Pel));
        }
      }
    }
  }
  else
  {
    const Pel *ptrSrc = getPredictorPtr( compID, bUseFilteredPredSamples );

    if ( uiDirMode == PLANAR_IDX )
    {
      xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType, format );
    }
    else
    {
      // Create the prediction
            TComDataCU *const pcCU              = rTu.getCU();
      const UInt              uiAbsPartIdx      = rTu.GetAbsPartIdxTU();
      const Bool              enableEdgeFilters = !(pcCU->isRDPCMEnabled(uiAbsPartIdx) && pcCU->getCUTransquantBypass(uiAbsPartIdx));

#if RExt__O0043_BEST_EFFORT_DECODING
      xPredIntraAng( g_bitDepthInStream[channelType], ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType, format, uiDirMode, bAbove, bLeft, enableEdgeFilters );
#else
      xPredIntraAng( g_bitDepth[channelType], ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType, format, uiDirMode, bAbove, bLeft, enableEdgeFilters );
#endif

      if(( uiDirMode == DC_IDX ) && bAbove && bLeft )
      {
        xDCPredFiltering( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType );
      }
    }
  }

}

/** Function for checking identical motion.
 * \param TComDataCU* pcCU
 * \param UInt PartAddr
 */
Bool TComPrediction::xCheckIdenticalMotion ( TComDataCU* pcCU, UInt PartAddr )
{
  if( pcCU->getSlice()->isInterB() && !pcCU->getSlice()->getPPS()->getWPBiPred() )
  {
    if( pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr) >= 0 && pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr) >= 0)
    {
      Int RefPOCL0 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr))->getPOC();
      Int RefPOCL1 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_1, pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr))->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcCU->getCUMvField(REF_PIC_LIST_0)->getMv(PartAddr) == pcCU->getCUMvField(REF_PIC_LIST_1)->getMv(PartAddr))
      {
        return true;
      }
    }
  }
  return false;
}


Void TComPrediction::intraBlockCopy ( TComDataCU* pcCU, TComYuv* pcYuvPred, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

  TComMv      cMv         = pcCU->getCUMvField( REF_PIC_LIST_INTRABC )->getMv( uiPartAddr );

  xPredIntraBCBlk( COMPONENT_Y, pcCU, pcCU->getPic()->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, pcYuvPred );

  if( pcYuvPred->getChromaFormat() != CHROMA_400 )
  {
    if( pcCU->getWidth(0) == 8 && pcCU->getPartitionSize(0) != SIZE_2Nx2N && pcYuvPred->getChromaFormat() != CHROMA_444 )
    {
      // the chroma PU will be smaller than 4x4, so join with neighbouring chroma PU(s) to form a bigger block
      // chroma PUs will use the luma MV from the bottom right most of the merged chroma PUs.
      UInt uiMvSrcAddr = ( pcYuvPred->getChromaFormat() == CHROMA_422 && iPartIdx < ( pcCU->getNumPartitions() >> 1 ) ? 1 : 3 );
      cMv = pcCU->getCUMvField( REF_PIC_LIST_INTRABC )->getMv( uiMvSrcAddr );
    }

    xPredIntraBCBlk( COMPONENT_Cb, pcCU, pcCU->getPic()->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, pcYuvPred );
    xPredIntraBCBlk( COMPONENT_Cr, pcCU, pcCU->getPic()->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, pcYuvPred );
  }

  return;
}

Void TComPrediction::motionCompensation ( TComDataCU* pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  if ( iPartIdx >= 0 )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
    return;
  }

  for ( iPartIdx = 0; iPartIdx < pcCU->getNumPartitions(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
  }
  return;
}

Void TComPrediction::xPredInterUni ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);

  for (UInt ch=COMPONENT_Y; ch<rpcYuvPred->getNumberValidComponents(); ch++)
    xPredInterBlk  (ComponentID(ch),  pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
}

Void TComPrediction::xPredInterBi ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[NUM_REF_PIC_LIST_01] = {-1, -1};

  for ( UInt refList = 0; refList < NUM_REF_PIC_LIST_01; refList++ )
  {
    RefPicList eRefPicList = (refList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[refList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[refList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[refList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[refList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) ||
           ( pcCU->getSlice()->getPPS()->getWPBiPred()    && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
      }
      else
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred()    && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[REF_PIC_LIST_0], &m_acYuvPred[REF_PIC_LIST_1], iRefIdx[REF_PIC_LIST_0], iRefIdx[REF_PIC_LIST_1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[REF_PIC_LIST_0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred );
  }
  else
  {
    xWeightedAverage( &m_acYuvPred[REF_PIC_LIST_0], &m_acYuvPred[REF_PIC_LIST_1], iRefIdx[REF_PIC_LIST_0], iRefIdx[REF_PIC_LIST_1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}

/**
 * \brief Generate motion-compensated block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */


Void TComPrediction::xPredInterBlk(const ComponentID compID, TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getStride(compID);
  Int     dstStride  = dstPic->getStride(compID);
  Int shiftHor=(2+refPic->getComponentScaleX(compID));
  Int shiftVer=(2+refPic->getComponentScaleY(compID));

  Int     refOffset  = (mv->getHor() >> shiftHor) + (mv->getVer() >> shiftVer) * refStride;

  Pel*    ref     = refPic->getAddr(compID, cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;

  Pel*    dst = dstPic->getAddr( compID, partAddr );

  Int     xFrac  = mv->getHor() & ((1<<shiftHor)-1);
  Int     yFrac  = mv->getVer() & ((1<<shiftVer)-1);
  UInt    cxWidth  = width  >> refPic->getComponentScaleX(compID);
  UInt    cxHeight = height >> refPic->getComponentScaleY(compID);

  const ChromaFormat chFmt = cu->getPic()->getChromaFormat();

  if ( yFrac == 0 )
  {
    m_if.filterHor(compID, ref, refStride, dst,  dstStride, cxWidth, cxHeight, xFrac, !bi, chFmt);
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVer(compID, ref, refStride, dst, dstStride, cxWidth, cxHeight, yFrac, true, !bi, chFmt);
  }
  else
  {
    Int   tmpStride = m_filteredBlockTmp[0].getStride(compID);
    Pel*  tmp       = m_filteredBlockTmp[0].getAddr(compID);

    const Int vFilterSize = isLuma(compID) ? NTAPS_LUMA : NTAPS_CHROMA;

    m_if.filterHor(compID, ref - ((vFilterSize>>1) -1)*refStride, refStride, tmp, tmpStride, cxWidth, cxHeight+vFilterSize-1, xFrac, false,      chFmt);
    m_if.filterVer(compID, tmp + ((vFilterSize>>1) -1)*tmpStride, tmpStride, dst, dstStride, cxWidth, cxHeight,               yFrac, false, !bi, chFmt);
  }
}

Void TComPrediction::xPredIntraBCBlk(const ComponentID compID, TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic )
{
  Int     refStride  = refPic->getStride(compID);
  Int     dstStride  = dstPic->getStride(compID);

  Int mvx = mv->getHor() >>  refPic->getComponentScaleX(compID);
  Int mvy = mv->getVer() >>  refPic->getComponentScaleY(compID);

  Int     refOffset  = mvx + mvy * refStride;

  Pel*    ref = refPic->getAddr( compID, cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  Pel*    dst = dstPic->getAddr( compID, partAddr );

  UInt    cxWidth  = width  >> refPic->getComponentScaleX(compID);
  UInt    cxHeight = height >> refPic->getComponentScaleY(compID);

  for (Int row = 0; row < cxHeight; row++)
  {
    for (Int col = 0; col < cxWidth; col++)
    {
      dst[col] = ref[col];
    }

    ref += refStride;
    dst += dstStride;
  }
}

Void TComPrediction::xWeightedAverage( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst )
{
  if( iRefIdx0 >= 0 && iRefIdx1 >= 0 )
  {
    rpcYuvDst->addAvg( pcYuvSrc0, pcYuvSrc1, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 >= 0 && iRefIdx1 <  0 )
  {
    pcYuvSrc0->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 <  0 && iRefIdx1 >= 0 )
  {
    pcYuvSrc1->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
}

// AMVP
Void TComPrediction::getMvPredAMVP( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred )
{
  AMVPInfo* pcAMVPInfo = pcCU->getCUMvField(eRefPicList)->getAMVPInfo();

  if( pcAMVPInfo->iN <= 1 )
  {
    rcMvPred = pcAMVPInfo->m_acMvCand[0];

    pcCU->setMVPIdxSubParts( 0, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    pcCU->setMVPNumSubParts( pcAMVPInfo->iN, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    return;
  }

  assert(pcCU->getMVPIdx(eRefPicList,uiPartAddr) >= 0);
  rcMvPred = pcAMVPInfo->m_acMvCand[pcCU->getMVPIdx(eRefPicList,uiPartAddr)];
  return;
}

/** Function for deriving planar intra prediction.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 *
 * This function derives the prediction samples for planar mode (intra coding).
 */
//NOTE: Bit-Limit - 24-bit source
Void TComPrediction::xPredIntraPlanar( const Pel* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height, ChannelType channelType, ChromaFormat format )
{
  assert(width <= height);

  Int leftColumn[MAX_CU_SIZE+1], topRow[MAX_CU_SIZE+1], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
  UInt shift1Dhor = g_aucConvertToBit[ width ] + 2;
  UInt shift1Dver = g_aucConvertToBit[ height ] + 2;

  // Get left and above reference column and row
  for(Int k=0;k<width+1;k++)
  {
    topRow[k] = pSrc[k-srcStride];
  }

  for (Int k=0; k < height+1; k++)
  {
    leftColumn[k] = pSrc[k*srcStride-1];
  }

  // Prepare intermediate variables used in interpolation
  Int bottomLeft = leftColumn[height];
  Int topRight   = topRow[width];

  for(Int k=0;k<width;k++)
  {
    bottomRow[k]  = bottomLeft - topRow[k];
    topRow[k]     <<= shift1Dver;
  }

  for(Int k=0;k<height;k++)
  {
    rightColumn[k]  = topRight - leftColumn[k];
    leftColumn[k]   <<= shift1Dhor;
  }

  const UInt topRowShift = 0;

  // Generate prediction signal
  for (Int y=0;y<height;y++)
  {
    Int horPred = leftColumn[y] + width;
    for (Int x=0;x<width;x++)
    {
      horPred += rightColumn[y];
      topRow[x] += bottomRow[x];

      // NOTE: RExt - The intermediate shift right could be rolled into the final shift right,
      //              thereby increasing the accuracy of the calculation
      // eg  rpDst[y*dstStride+x] = ( (horPred<<topRowShift) + topRow[x] ) >> (shift1Dver+1);
      Int vertPred = ((topRow[x] + topRowShift)>>topRowShift);
      rpDst[y*dstStride+x] = ( horPred + vertPred ) >> (shift1Dhor+1);
    }
  }
}

/** Function for filtering intra DC predictor.
 * \param pSrc pointer to reconstructed sample array
 * \param iSrcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param iWidth the width of the block
 * \param iHeight the height of the block
 *
 * This function performs filtering left and top edges of the prediction samples for DC mode (intra coding).
 */
Void TComPrediction::xDCPredFiltering( const Pel* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight, ChannelType channelType )
{
  Pel* pDst = rpDst;
  Int x, y, iDstStride2, iSrcStride2;

  if (isLuma(channelType) && (iWidth <= MAXIMUM_INTRA_FILTERED_WIDTH) && (iHeight <= MAXIMUM_INTRA_FILTERED_HEIGHT))
  {
    //top-left
    pDst[0] = (Pel)((pSrc[-iSrcStride] + pSrc[-1] + 2 * pDst[0] + 2) >> 2);

    //top row (vertical filter)
    for ( x = 1; x < iWidth; x++ )
    {
      pDst[x] = (Pel)((pSrc[x - iSrcStride] +  3 * pDst[x] + 2) >> 2);
    }

    //left column (horizontal filter)
    for ( y = 1, iDstStride2 = iDstStride, iSrcStride2 = iSrcStride-1; y < iHeight; y++, iDstStride2+=iDstStride, iSrcStride2+=iSrcStride )
    {
      pDst[iDstStride2] = (Pel)((pSrc[iSrcStride2] + 3 * pDst[iDstStride2] + 2) >> 2);
    }
  }

  return;
}

/* Static member function */
Bool TComPrediction::UseDPCMForFirstPassIntraEstimation(TComTU &rTu, const UInt uiDirMode)
{
  return (rTu.getCU()->isRDPCMEnabled(rTu.GetAbsPartIdxTU()) ) &&
          rTu.getCU()->getCUTransquantBypass(rTu.GetAbsPartIdxTU()) &&
          (uiDirMode==HOR_IDX || uiDirMode==VER_IDX);
}

#if SCM__R0348_PALETTE_MODE
Void TComPrediction::deriveRun(TComDataCU* pcCU, Pel* pOrg[3],  Pel *pPalette[3],  Pel* pValue, UChar* pSPoint,
                               Pel** paPixelValue, Pel ** paRecoValue, TCoeff* pRun,
                               UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{
  UInt uiTotal = uiHeight * uiWidth, uiIdx = 0;
  UInt uiStartPos = 0,  uiRun = 0, uiCopyRun = 0;
  Int iTemp = 0;
  UInt uiStride = uiWidth;
  UInt uiTraIdx;  //unified position variable (raster scan)

  //Test Run
  while (uiIdx < uiTotal)
  {
    uiStartPos = uiIdx;
    uiRun = 0;

    Bool RunValid = calLeftRun(pValue, pSPoint, uiStartPos, uiTotal, uiRun);
    uiCopyRun = 0;
    Bool CopyValid = calAboveRun(pValue, pSPoint, uiWidth, uiStartPos, uiTotal, uiCopyRun);
    uiTraIdx = m_puiScanOrder[uiIdx];    //unified position variable (raster scan)

    if (CopyValid == 0 && RunValid == 0)
    {
      pSPoint[uiTraIdx] = PLT_ESCAPE;
      calcPixelPred(pcCU, pOrg, pPalette, pValue, paPixelValue, paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiTraIdx);
      uiIdx++;
    }
    else
    {
      if ( CopyValid && (uiCopyRun + 2 > uiRun))
      {
        pSPoint[uiTraIdx] = PLT_RUN_ABOVE;
        pRun[uiTraIdx]  = uiCopyRun-1;
        uiIdx++;

        iTemp = uiCopyRun - 1;
        while (iTemp > 0)
        {
          uiTraIdx = m_puiScanOrder[uiIdx];  //unified position variable (raster scan)
          pValue[uiTraIdx] = pValue[uiTraIdx - uiStride];
          pSPoint[uiTraIdx] = PLT_RUN_ABOVE;
          uiIdx++;

          iTemp--;
        }
      }
      else
      {
        pSPoint[uiTraIdx] = PLT_RUN_LEFT;
        pRun[uiTraIdx] = uiRun;
        uiIdx++;
        iTemp = uiRun;
        while (iTemp > 0)
        {
          uiTraIdx = m_puiScanOrder[uiIdx];  //unified position variable (raster scan)
          pValue[uiTraIdx] = pValue[m_puiScanOrder[uiIdx - 1]];
          pSPoint[uiTraIdx] = PLT_RUN_LEFT;
          uiIdx++;
          iTemp--;
        }
      }
    }
  }
  assert (uiIdx == uiTotal);
}

#if IDX_1D_SEARCH || DYNAMIC_2D_SEARCH
Void TComPrediction::deriveMatchPair(TComDataCU* pcCU, Pel* pOrg [3],  Pel *pPalette [3],  Pel* pValue, UChar* pSPoint, Pel** paPixelValue, Pel ** paRecoValue, TCoeff* pDistLen, UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiPLTSize)
{
  UInt uiNumLCUHorz = pcCU->getPic()->getFrameWidthInCU();
  UInt uiCUAddr = pcCU->getAddr();
  UInt uiCUAddrHorz = uiCUAddr % uiNumLCUHorz;
  Bool bUse2D = false;

#if DYNAMIC_2D_SEARCH
  bUse2D = ( uiWidth > 32 );

  if(uiCUAddrHorz == 0)
  {
    bUse2D = false;
  }
#endif

  ////////////////////////////////////////////////////
  // start 1d/2d search
  ////////////////////////////////////////////////////

  Bool b2DSearchWithinCurCTU = bUse2D;

  Int zorder = pcCU->getZorderIdxInCU();
  UInt uiX0 = g_auiRasterToPelX[ g_auiZscanToRaster[zorder] ];
  UInt uiY0 = g_auiRasterToPelY[ g_auiZscanToRaster[zorder] ];

  // 2D Reference IdxMap & 2D Hash & RefCodedMap
  static UInt p2DHashDictLast[256];
  static UInt p2DHashDict[(D2D_LEFTWIDTH+1)*(D2D_TOPHEIGHT+1)*64*64];
  static UChar p2DHash[(D2D_LEFTWIDTH+1)*(D2D_TOPHEIGHT+1)*64*64];

  static UChar pbRefCodedMap[(D2D_LEFTWIDTH+1)*(D2D_TOPHEIGHT+1)*64*64];
  static UChar pRefIdx[(D2D_LEFTWIDTH+1)*(D2D_TOPHEIGHT+1)*64*64];

  // 1D IdxMap & 1D Hash & CodedMap
  static Pel  p1DHashDict[3][64*64];
  static UChar pbCodedMap[64*64];
  static UChar pIdx[64*64];

  UInt uiLcuWidth = 0, uiLcuHeight = 0;
  UInt uiRefWidth = 0, uiRefHeight = 0;

  if( bUse2D == 1 ) 
  {
    uiLcuWidth  = g_uiMaxCUWidth;
    uiLcuHeight = g_uiMaxCUHeight;

    UInt uiStride = pcCU->getPic()->getPicYuvRec()->getStride( (ComponentID) 0 );
    UInt uiPicWidth = pcCU->getPic()->getPicYuvRec()->getWidth( (ComponentID) 0 );
    UInt uiPicHeight = pcCU->getPic()->getPicYuvRec()->getHeight( (ComponentID) 0 );
    UInt uiNumLCUVert = pcCU->getPic()->getFrameHeightInCU();
    UInt uiCUAddrVert = uiCUAddr / uiNumLCUHorz;

    if( uiCUAddrHorz+1 == uiNumLCUHorz )
    {
      uiLcuWidth = uiPicWidth - uiCUAddrHorz*uiLcuWidth;
    }
    if( uiCUAddrVert+1 == uiNumLCUVert )
    {
      uiLcuHeight = uiPicHeight - uiCUAddrVert*uiLcuHeight;
    }

    uiRefWidth = uiLcuWidth + pcCU->getIntraBCSearchAreaWidth( D2D_LEFTWIDTH ); 
    uiRefHeight = uiLcuHeight + pcCU->getIntraBCSearchAreaHeight( D2D_TOPHEIGHT );

    Pel *paRec[3];  
    for (UInt ch=0; ch<3; ch++)
    {
      paRec[ch] = pcCU->getPic()->getPicYuvRec()->getAddr( (ComponentID) ch, pcCU->getAddr() );
      paRec[ch] -= (uiRefHeight - uiLcuHeight) * uiStride + (uiRefWidth - uiLcuWidth);
    }

    // generate pRefIdx[] from search window
    deriveColorIdxMap( pcCU, paRec, pPalette, pRefIdx, uiRefWidth, uiRefHeight, uiStride, uiPLTSize );

    // update pRefIdx[] with pcIndexBlock[]
    Pel *pcIndexBlock = m_cIndexBlock;
    UInt uiRefIdx = (uiRefHeight - uiLcuHeight + uiY0)*(uiRefWidth) + (uiRefWidth - uiLcuWidth + uiX0);
    for(UInt y = 0; y < uiHeight; ++y)
    {
      for(UInt x = 0; x < uiWidth; ++x)
      {
        pRefIdx[uiRefIdx + x + y*uiRefWidth] = (UChar)pcIndexBlock[x + y*uiWidth];
      }
    }

    // 2D hash
    derive2DHash( pRefIdx, uiRefWidth, uiRefHeight, p2DHash, p2DHashDict, p2DHashDictLast );

    // RefCodedMap
    memset( pbRefCodedMap,   0, sizeof(UChar)*(D2D_LEFTWIDTH+1)*(D2D_TOPHEIGHT+1)*64*64 );
    memset( pbRefCodedMap,  -1, sizeof(UChar)*(uiRefHeight - uiLcuHeight)*(uiRefWidth) );

    for(UInt y = uiRefHeight - uiLcuHeight; y < uiRefHeight; ++y)
    {
      for(UInt x = 0; x < uiRefWidth - uiLcuWidth; ++x)
      {
        pbRefCodedMap[x + y*uiRefWidth] = -1;
      }
    }

    if( b2DSearchWithinCurCTU == 1 )
    {
      Int width = g_uiMaxCUWidth>>(g_uiMaxCUDepth-1);
      Int height = g_uiMaxCUHeight>>(g_uiMaxCUDepth-1);
      UInt uiRefIdx2 = (uiRefHeight - uiLcuHeight)*(uiRefWidth) + (uiRefWidth - uiLcuWidth);

      for(Int n = 0; n < zorder; n+=4)
      {
        UInt x0 = g_auiRasterToPelX[ g_auiZscanToRaster[n] ];
        UInt y0 = g_auiRasterToPelY[ g_auiZscanToRaster[n] ];
        UInt x1 = ( x0+width <= uiLcuWidth ) ? x0+width : uiLcuWidth;
        UInt y1 = ( y0+height <= uiLcuHeight ) ? y0+height : uiLcuHeight;

        for(UInt y = y0; y < y1; ++y)
        {
          for(UInt x = x0; x < x1; ++x)
          {
            pbRefCodedMap[uiRefIdx2 + x + y*uiRefWidth] = -1;
          }
        }
      }	
    }
  }

  // update pIdx[] with pcIndexBlock[]
  Pel *pcIndexBlock = m_cIndexBlock;
  for(UInt y = 0; y < uiHeight; ++y)
  {
    for(UInt x = 0; x < uiWidth; ++x)
    {
      pIdx[x + y*uiWidth] = (UChar)pcIndexBlock[x + y*uiWidth];
    }
  }

  // CodedMap
  memset( pbCodedMap, 0, sizeof(UChar)*64*64 );

  // copy index to Y buffer
  UInt uiIdx = 0, uiTraIdx;
  const UInt uiTotal = uiWidth*uiHeight;
  memcpy( pValue, pcIndexBlock, sizeof(Pel)*uiTotal );

  ////////////////////////////////////////////////////
  // start 1d/2d search
  ////////////////////////////////////////////////////

  {
    while (uiIdx < uiTotal)
    {
      // 1D search
      Int smax1DLen = 0;
      Int max1DLen  = 0;
      Int d1Dist    = -1;

      Derive1DSearch(pIdx, uiWidth, uiHeight, uiTotal, pbCodedMap, p1DHashDict, uiIdx, max1DLen, d1Dist, bUse2D);

      smax1DLen = max1DLen;

      // 2D search
      Int max2DWidth  = 0;
      Int max2DHeight = 0;
      Int max2DSize   = 0;
      Int d2Dist      = -1;

      if( bUse2D == 1 ) 
      {
        Derive2DSearch( pcCU, pIdx, pbCodedMap, uiIdx, uiWidth, uiHeight, uiX0, uiY0, uiLcuWidth, uiLcuHeight, 
                        pRefIdx, pbRefCodedMap, uiRefWidth, uiRefHeight, p2DHash, p2DHashDict, p2DHashDictLast, 
                        max2DSize, max2DWidth, max2DHeight, d2Dist);

        for(Int n = 0; n < max1DLen; ++n)
        {
          // remove already coded location from len
          if( pbCodedMap[m_puiScanOrder[uiIdx+n]] != 0 )
          {
            --smax1DLen;
          }
        }
      }

      // 1D vs 2D
      if( max2DSize <= (smax1DLen*1.1) || max2DWidth < MIN_D2D_WIDTH || max2DHeight < MIN_D2D_HEIGHT )       // 1D is better, or 2D smaller than MIN_D2D_WIDTHxMIN_D2D_HEIGHT
      {
        if( bUse2D == 1 )
        {
          for(Int n = 0; n < ((smax1DLen < 1) ? 1 : max1DLen); ++n)
          {
            uiTraIdx = m_puiScanOrder[uiIdx+n];

            if( pbCodedMap[uiTraIdx] == 0 )
            {
              pbCodedMap[uiTraIdx] = -1;
              
              if( b2DSearchWithinCurCTU == 1 )
              {
                UInt uiRefIdx = ( uiRefHeight - uiLcuHeight + ( uiY0 + ( uiTraIdx / uiWidth ) ) ) * (uiRefWidth) + (uiRefWidth - uiLcuWidth) + ( uiX0 + ( uiTraIdx % uiWidth ) );
                pbRefCodedMap[uiRefIdx] = -1;
              }
            }
          }
        }

        uiTraIdx = m_puiScanOrder[uiIdx];

        if(smax1DLen < 1)
        {
          pSPoint[uiTraIdx] = PLT_RUN_LEFT;
          pDistLen[uiTraIdx] = (0 << 16) | (1);

          uiIdx++;
        }
        else
        {
          pSPoint[uiTraIdx] = ( d1Dist <= uiTraIdx % uiWidth ) ? PLT_RUN_LEFT : PLT_RUN_ABOVE;
          pDistLen[uiTraIdx] = (d1Dist << 16) | (max1DLen);

          uiIdx = uiIdx + max1DLen;
        }
      }
      else                                            // 2D is better
      {
        uiTraIdx = m_puiScanOrder[uiIdx];
        Bool bTraverse = ( (uiTraIdx / uiWidth) & 1 );

        pSPoint[uiTraIdx] = ( uiRefHeight - d2Dist/uiRefWidth == uiHeight - uiTraIdx/uiWidth ) ? PLT_D2D | PLT_RUN_LEFT : PLT_D2D | PLT_RUN_ABOVE;
        pDistLen[uiTraIdx] = (d2Dist << 16) | (max2DWidth << 8) | (max2DHeight);
      
        UInt uiRefIdx = (uiRefHeight - uiLcuHeight + (uiY0 + (uiTraIdx / uiWidth)))*(uiRefWidth) + (uiRefWidth - uiLcuWidth) + (uiX0 + (uiTraIdx % uiWidth));

        for(Int y = 0; y < max2DHeight; ++y)
        {
          for(Int x = 0; x < max2DWidth; ++x)
          {
            Int traX = (bTraverse) ? -x : x;

            pbCodedMap[uiTraIdx+traX+y*uiWidth] = -1;
          
            if( b2DSearchWithinCurCTU == 1 )
            {
              pbRefCodedMap[uiRefIdx+traX+y*uiRefWidth] = -1;
            }
          }
        }
      
        uiIdx = uiIdx + max2DWidth;
      }

      // skip already coded location
      while( (uiIdx < uiTotal) && (pbCodedMap[m_puiScanOrder[uiIdx]] != 0) )
      {
        ++uiIdx;
      }
    }

    assert( uiIdx == uiTotal );
  }

  for( uiTraIdx = 0; uiTraIdx < uiTotal; uiTraIdx++ )
  {
    if( pValue[uiTraIdx] == uiPLTSize )
    {
      pSPoint[uiTraIdx] |= PLT_ESCAPE; 
      calcPixelPred(pcCU, pOrg, pPalette, pValue, paPixelValue, paRecoValue, uiWidth, uiHeight, uiStrideOrg, uiTraIdx );
    }
  }
}

Void TComPrediction::Derive1DSearch(UChar* pIdx, UInt uiWidth, UInt uiHeight, UInt uiTotal, UChar* pbCodedMap, Pel p1DHashDict[3][64*64], UInt uiIdx, Int& max1DLen, Int& d1Dist, Bool bUse2D)
{
  Int smax1DLen = max1DLen;

  // dist = 0
  {
    Int dist = 0;
    UChar first = pIdx[m_puiScanOrder[uiIdx]];

    {
      Int len = 0;
      
      for (len = 0; len < (uiTotal-uiIdx); ++len )
      {
        UInt uiTraIdx = m_puiScanOrder[uiIdx+len];
        UChar cmp = ( first ^ pIdx[uiTraIdx] ) & ( ~pbCodedMap[uiTraIdx] );
        
        if( cmp )
        {
          break;
        }
      }
      
      Int slen = len;

      if( bUse2D == 1 )
      {
        for(Int n = 0; n < slen; ++n)
        {
          // remove already coded location from len
          if( pbCodedMap[m_puiScanOrder[uiIdx+n]] == (UChar)-1 )
          {
            --slen;
          }
        }
      }

      // better to change with R-D decision
      if( slen > smax1DLen )
      {
        smax1DLen = slen;
        max1DLen  = len;
        d1Dist    = dist;
      }
    }
  }
  
  // dist != 0
  Int refIdx = m_puiScanOrder[uiIdx]-uiWidth;
  
  for( ; refIdx >= 0; refIdx -= uiWidth )
  {
    Int dist =  m_puiScanOrder[uiIdx] - refIdx;

    {
      Int len = 0;
      
      for (len = 0; len < (uiTotal-uiIdx); ++len )
      {
        UInt uiTraIdx = m_puiScanOrder[uiIdx+len];
        UChar cmp = ( pIdx[uiTraIdx-dist] ^ pIdx[uiTraIdx] ) & ( ~pbCodedMap[uiTraIdx] );
        
        if( cmp )
        {
          break;
        }
      }
      
      Int slen = len;

      if( bUse2D == 1 )
      {
        for(Int n = 0; n < slen; ++n)
        {
          // remove already coded location from len
          if( pbCodedMap[m_puiScanOrder[uiIdx+n]] != 0 )
          {
            --slen;
          }
        }
      }

      // better to change with R-D decision
      if( slen > smax1DLen ||  (dist == uiWidth && slen == smax1DLen) || (d1Dist == 0 && slen != 0 && slen + 2 > smax1DLen) )
      {
        smax1DLen = slen;
        max1DLen  = len;
        d1Dist    = dist;
      }
    }

    break;
  }
}

Void TComPrediction::Derive2DSearch(TComDataCU* pcCU, UChar* pIdx, UChar* pbCodedMap, UInt uiIdx, UInt uiWidth, UInt uiHeight, UInt uiX0, UInt uiY0, UInt uiLcuWidth, UInt uiLcuHeight, 
                                    UChar* pRefIdx, UChar* pbRefCodedMap, UInt uiRefWidth, UInt uiRefHeight, UChar* p2DHash, UInt* p2DHashDict, UInt* p2DHashDictLast, 
                                    Int& max2DSize, Int& max2DWidth, Int& max2DHeight, Int& d2Dist)
{
  UInt uiTraIdx = m_puiScanOrder[uiIdx];
  Bool bTraverse = ( (uiTraIdx / uiWidth) & 1 );

  if( bTraverse )
  {
    Int cw0 = 1 + uiTraIdx % uiWidth;
    Int ch0 = uiHeight - uiTraIdx / uiWidth;
  
    if( (cw0 >= MIN_D2D_WIDTH) && (ch0 >= MIN_D2D_HEIGHT) ) 
    {
      // dist = 0
      UInt uiIdx0 = (uiRefHeight - uiLcuHeight + (uiY0 + (uiTraIdx / uiWidth)))*(uiRefWidth) + (uiRefWidth - uiLcuWidth) + (uiX0 + (uiTraIdx % uiWidth));

      {
        {
          Int minw = cw0;
          Int minh = ch0;
        
          UInt64 curIdx = pIdx[uiTraIdx];
          UInt64 first = (curIdx<<56) | (curIdx<<48) | (curIdx<<40) | (curIdx<<32) | (curIdx<<24) | (curIdx<<16) | (curIdx<<8) | (curIdx);
          Int d2Width = MAX_INT;
          Int d2Size = 0;
        
          for ( Int y = 0; y < minh; ++y )
          {
            Int x = 0;
          
            for( x = 0; x < (minw&0xf8); x+=8 )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiTraIdx-x-7+y*uiWidth) ) ) & ( first ^ ( *(UInt64*)(pIdx+uiTraIdx-x-7+y*uiWidth) ) );
            
              if( cmp )
              {
                x += (cmp & 0xff00000000000000) ? 0 : (cmp & 0xff000000000000) ? 1 : (cmp & 0xff0000000000) ? 2 : (cmp & 0xff00000000) ? 3 : (cmp & 0xff000000) ? 4 : (cmp & 0xff0000) ? 5 : (cmp & 0xff00) ? 6 : 7;
                break;
              }
            }
            if( x == (minw&0xf8) && (minw&7) )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiTraIdx-x-7+y*uiWidth) ) ) & ( first ^ ( *(UInt64*)(pIdx+uiTraIdx-x-7+y*uiWidth) ) );
              cmp &= ~( ( ((UInt64)1) << ((8-(minw&7))<<3) ) - 1 );
            
              if( cmp )
              {
                x += (cmp & 0xff00000000000000) ? 0 : (cmp & 0xff000000000000) ? 1 : (cmp & 0xff0000000000) ? 2 : (cmp & 0xff00000000) ? 3 : (cmp & 0xff000000) ? 4 : (cmp & 0xff0000) ? 5 : (cmp & 0xff00) ? 6 : 7;
              }
              else
              {
                x = minw;
              }
            }
          
            // ignore any 2d block if width < MIN_2D_WIDTH
            if( x < MIN_D2D_WIDTH )   
              break;
          
            d2Width = ( d2Width < x ) ? d2Width : x;
            d2Size = d2Width * (y+1);
            minw = d2Width;
          
            if ( d2Size > max2DSize )       
            {
              max2DSize   = d2Size;
              max2DWidth  = d2Width;
              max2DHeight = y+1;
              d2Dist      = uiIdx0;
            }
          }
        }
      }

      // dist != 0
      for( Int refIdx = p2DHashDictLast[p2DHash[uiIdx0-MIN_D2D_WIDTH]]+MIN_D2D_WIDTH; refIdx != -1+MIN_D2D_WIDTH; refIdx = p2DHashDict[refIdx-MIN_D2D_WIDTH]+MIN_D2D_WIDTH )
      {
        Int rx = refIdx % uiRefWidth;
        Int ry = refIdx / uiRefWidth;
      
        if ( ( pRefIdx[refIdx] == pIdx[uiTraIdx] ) && ( pbRefCodedMap[refIdx] != 0 ) )
        {
          Int rw0 = 1 + rx;
          Int rh0 = uiRefHeight - ry;
        
          if( rw0 < MIN_D2D_WIDTH || rh0 < MIN_D2D_HEIGHT )
              continue;
        
          Int minw = (rw0 < cw0) ? rw0 : cw0;
          Int minh = (rh0 < ch0) ? rh0 : ch0;
        
          Int d2Width = MAX_INT;
          Int d2Size = 0;
        
          for ( Int y = 0; y < minh; ++y )
          {
            Int x = 0;
          
            for( x = 0; x < (minw&0xf8); x+=8 )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiTraIdx-x-7+y*uiWidth) ) ) & ( ( ~( *(UInt64*)(pbRefCodedMap+refIdx-x-7+y*uiRefWidth) ) ) | ( ( *(UInt64*)(pRefIdx+refIdx-x-7+y*uiRefWidth) ) ^ ( *(UInt64*)(pIdx+uiTraIdx-x-7+y*uiWidth) ) ) );
            
              if( cmp )
              {
                x += (cmp & 0xff00000000000000) ? 0 : (cmp & 0xff000000000000) ? 1 : (cmp & 0xff0000000000) ? 2 : (cmp & 0xff00000000) ? 3 : (cmp & 0xff000000) ? 4 : (cmp & 0xff0000) ? 5 : (cmp & 0xff00) ? 6 : 7;
                break;
              }
            }
            if( x == (minw&0xf8) && (minw&7) )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiTraIdx-x-7+y*uiWidth) ) ) & ( ( ~( *(UInt64*)(pbRefCodedMap+refIdx-x-7+y*uiRefWidth) ) ) | ( ( *(UInt64*)(pRefIdx+refIdx-x-7+y*uiRefWidth) ) ^ ( *(UInt64*)(pIdx+uiTraIdx-x-7+y*uiWidth) ) ) );
              cmp &= ~( ( ((UInt64)1) << ((8-(minw&7))<<3) ) - 1 );
            
              if( cmp )
              {
                x += (cmp & 0xff00000000000000) ? 0 : (cmp & 0xff000000000000) ? 1 : (cmp & 0xff0000000000) ? 2 : (cmp & 0xff00000000) ? 3 : (cmp & 0xff000000) ? 4 : (cmp & 0xff0000) ? 5 : (cmp & 0xff00) ? 6 : 7;
              }
              else
              {
                x = minw;
              }
            }
          
            // ignore any 2d block if width < MIN_2D_WIDTH
            if( x < MIN_D2D_WIDTH )   
              break;
          
            d2Width = ( d2Width < x ) ? d2Width : x;
            d2Size = d2Width * (y+1);
            minw = d2Width;
          
            if ( d2Size > max2DSize || (d2Dist == 0 && d2Size + 2 > max2DSize) )       
            {
              max2DSize   = d2Size;
              max2DWidth  = d2Width;
              max2DHeight = y+1;
              d2Dist      = refIdx;
            }
          }
        }
      }
    }
  }
  else
  {
    Int cw0 = uiWidth - uiIdx % uiWidth;
    Int ch0 = uiHeight - uiIdx / uiWidth;
  
    if( (cw0 >= MIN_D2D_WIDTH) && (ch0 >= MIN_D2D_HEIGHT) ) 
    {
      // dist = 0
      UInt uiIdx0 = (uiRefHeight - uiLcuHeight + (uiY0 + (uiIdx / uiWidth)))*(uiRefWidth) + (uiRefWidth - uiLcuWidth) + (uiX0 + (uiIdx % uiWidth));

      {
        {
          Int minw = cw0;
          Int minh = ch0;
        
          UInt64 curIdx = pIdx[uiIdx];
          UInt64 first = (curIdx<<56) | (curIdx<<48) | (curIdx<<40) | (curIdx<<32) | (curIdx<<24) | (curIdx<<16) | (curIdx<<8) | (curIdx);
          Int d2Width = MAX_INT;
          Int d2Size = 0;
        
          for ( Int y = 0; y < minh; ++y )
          {
            Int x = 0;
          
            for( x = 0; x < (minw&0xf8); x+=8 )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiIdx+x+y*uiWidth) ) ) & ( first ^ ( *(UInt64*)(pIdx+uiIdx+x+y*uiWidth) ) );
            
              if( cmp )
              {
                x += (cmp & 0xff) ? 0 : (cmp & 0xff00) ? 1 : (cmp & 0xff0000) ? 2 : (cmp & 0xff000000) ? 3 : (cmp & 0xff00000000) ? 4 : (cmp & 0xff0000000000) ? 5 : (cmp & 0xff000000000000) ? 6 : 7;
                break;
              }
            }
            if( x == (minw&0xf8) && (minw&7) )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiIdx+x+y*uiWidth) ) ) & ( first ^ ( *(UInt64*)(pIdx+uiIdx+x+y*uiWidth) ) );
              cmp &= ( ( ((UInt64)1) << ((minw&7)<<3) ) - 1 );
            
              if( cmp )
              {
                x += (cmp & 0xff) ? 0 : (cmp & 0xff00) ? 1 : (cmp & 0xff0000) ? 2 : (cmp & 0xff000000) ? 3 : (cmp & 0xff00000000) ? 4 : (cmp & 0xff0000000000) ? 5 : (cmp & 0xff000000000000) ? 6 : 7;
              }
              else
              {
                x = minw;
              }
            }
          
            // ignore any 2d block if width < MIN_2D_WIDTH
            if( x < MIN_D2D_WIDTH )   
              break;
          
            d2Width = ( d2Width < x ) ? d2Width : x;
            d2Size = d2Width * (y+1);
            minw = d2Width;
          
            if ( d2Size > max2DSize )       
            {
              max2DSize   = d2Size;
              max2DWidth  = d2Width;
              max2DHeight = y+1;
              d2Dist      = uiIdx0;
            }
          }
        }
      }

      // dist != 0
      for( Int refIdx = p2DHashDictLast[p2DHash[uiIdx0]]; refIdx != -1; refIdx = p2DHashDict[refIdx] )
      {
        Int rx = refIdx % uiRefWidth;
        Int ry = refIdx / uiRefWidth;
      
        if ( ( pRefIdx[refIdx] == pIdx[uiIdx] ) && ( pbRefCodedMap[refIdx] != 0 ) )
        {
          Int rw0 = uiRefWidth - rx;
          Int rh0 = uiRefHeight - ry;
        
          if( rw0 < MIN_D2D_WIDTH || rh0 < MIN_D2D_HEIGHT )
              continue;
        
          Int minw = (rw0 < cw0) ? rw0 : cw0;
          Int minh = (rh0 < ch0) ? rh0 : ch0;
        
          Int d2Width = MAX_INT;
          Int d2Size = 0;
        
          for ( Int y = 0; y < minh; ++y )
          {
            Int x = 0;
          
            for( x = 0; x < (minw&0xf8); x+=8 )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiIdx+x+y*uiWidth) ) ) & ( ( ~( *(UInt64*)(pbRefCodedMap+refIdx+x+y*uiRefWidth) ) ) | ( ( *(UInt64*)(pRefIdx+refIdx+x+y*uiRefWidth) ) ^ ( *(UInt64*)(pIdx+uiIdx+x+y*uiWidth) ) ) );
            
              if( cmp )
              {
                x += (cmp & 0xff) ? 0 : (cmp & 0xff00) ? 1 : (cmp & 0xff0000) ? 2 : (cmp & 0xff000000) ? 3 : (cmp & 0xff00000000) ? 4 : (cmp & 0xff0000000000) ? 5 : (cmp & 0xff000000000000) ? 6 : 7;
                break;
              }
            }
            if( x == (minw&0xf8) && (minw&7) )
            {
              Int64 cmp = ( ~( *(UInt64*)(pbCodedMap+uiIdx+x+y*uiWidth) ) ) & ( ( ~( *(UInt64*)(pbRefCodedMap+refIdx+x+y*uiRefWidth) ) ) | ( ( *(UInt64*)(pRefIdx+refIdx+x+y*uiRefWidth) ) ^ ( *(UInt64*)(pIdx+uiIdx+x+y*uiWidth) ) ) );
              cmp &= ( ( ((UInt64)1) << ((minw&7)<<3) ) - 1 );
            
              if( cmp )
              {
                x += (cmp & 0xff) ? 0 : (cmp & 0xff00) ? 1 : (cmp & 0xff0000) ? 2 : (cmp & 0xff000000) ? 3 : (cmp & 0xff00000000) ? 4 : (cmp & 0xff0000000000) ? 5 : (cmp & 0xff000000000000) ? 6 : 7;
              }
              else
              {
                x = minw;
              }
            }
          
            // ignore any 2d block if width < MIN_2D_WIDTH
            if( x < MIN_D2D_WIDTH )   
              break;
          
            d2Width = ( d2Width < x ) ? d2Width : x;
            d2Size = d2Width * (y+1);
            minw = d2Width;
          
            if ( d2Size > max2DSize || (d2Dist == 0 && d2Size + 2 > max2DSize) )
            {
              max2DSize   = d2Size;
              max2DWidth  = d2Width;
              max2DHeight = y+1;
              d2Dist      = refIdx;
            }
          }
        }
      }
    }
  }
}

Void TComPrediction::deriveColorIdxMap(TComDataCU* pcCU, Pel* pOrg [3],  Pel *pPalette[3], UChar* pIdx, UInt uiWidth, UInt uiHeight, UInt uiStride, UInt uiColorTableSize)
{
  UInt uiYCbCrShift = pcCU->getPic()->getComponentScaleY(COMPONENT_Cb);
  UInt uiXCbCrShift = pcCU->getPic()->getComponentScaleX(COMPONENT_Cb);
  UInt uiStrideCbCr = uiStride >> uiXCbCrShift;

  map<UInt, UChar> mColorTable;
  map<UInt, UChar>::iterator it;

  Bool bLossless = pcCU->isLosslessCoded( 0 );
  Int iErrorLimit = bLossless ? 0 : 3 * getPLTErrLimit();
  
  for( UInt uiY = 0; uiY < uiHeight; ++uiY)
  {
    for( UInt uiX = 0; uiX < uiWidth; ++uiX)
    {
      UInt uiYCbCr = uiY>>uiYCbCrShift;
      UInt uiXCbCr = uiX>>uiXCbCrShift;
      UInt uiYIdxRaster    = uiY*uiStride + uiX;
      UInt uiCbCrIdxRaster = uiYCbCr* uiStrideCbCr + uiXCbCr;
      UInt uiIdx = uiX + uiY * uiWidth;
      pIdx[uiIdx] = deriveColorIdx(pcCU, pPalette, uiColorTableSize, pOrg[0][uiYIdxRaster], pOrg[1][uiCbCrIdxRaster], pOrg[2][uiCbCrIdxRaster], iErrorLimit, mColorTable );
    }
  }

  return;
}

UChar TComPrediction::deriveColorIdx(TComDataCU* pcCU, Pel* pPalette[3], UInt uiColorTableSize, Pel pelComp0, Pel pelComp1, Pel pelComp2, Int iErrorLimit, map<UInt, UChar>& mColorTable)
{
  UInt uiColor = (pelComp0<<16)|(pelComp1<<8)|(pelComp2);
  map<UInt, UChar>::iterator it = mColorTable.find( uiColor );
  if( it != mColorTable.end() )
  {
    return it->second;
  }

  UInt uiMinError = MAX_UINT;
  UChar uiBestIdx = 0;
  for( UChar uiColorIdx = 0; uiColorIdx < uiColorTableSize; ++uiColorIdx )
  {
    UInt uiAbsError = abs( pPalette[0][uiColorIdx]-pelComp0 )+abs( pPalette[1][uiColorIdx]-pelComp1 )+abs( pPalette[2][uiColorIdx]-pelComp2 );

    if( uiAbsError < uiMinError )
    {
      uiBestIdx = uiColorIdx;
      uiMinError = uiAbsError;
      if (uiAbsError == 0)
      {
        break;
      }
    }
  }

  if( uiMinError > iErrorLimit )
  {
      uiBestIdx = uiColorTableSize;
  }

  mColorTable.insert( std::make_pair( uiColor, uiBestIdx ) );

  //assert ( uiBestIdx != (UChar)-1 );
  return uiBestIdx;
}

UChar TComPrediction::deriveColorIdx (TComDataCU* pcCU, Pel* pOrg [3],  Pel *pPalette[3], UInt uiIdx, UInt uiWidth, UInt uiHeight, UInt uiStride, UInt uiStrideCbCr, UInt uiXCbCrShift, UInt uiYCbCrShift, UInt uiColorTableSize, map<UInt, UChar> &mColorTable)
{
  UInt uiY = uiIdx/uiWidth;
  UInt uiX = uiIdx%uiWidth;
  UInt uiYCbCr = uiY>>uiYCbCrShift;
  UInt uiXCbCr = uiX>>uiXCbCrShift;
  UInt uiYIdxRaster    = uiY*uiStride + uiX;
  UInt uiCbCrIdxRaster = uiYCbCr* uiStrideCbCr + uiXCbCr;

  Bool bLossless = pcCU->isLosslessCoded( 0 );
  Int iErrorLimit = bLossless ? 0 : 3 * g_uhPLTQuant[Int(pcCU->getQP(0))];

  return deriveColorIdx(pcCU, pPalette, uiColorTableSize, pOrg[0][uiYIdxRaster], pOrg[1][uiCbCrIdxRaster], pOrg[2][uiCbCrIdxRaster], iErrorLimit, mColorTable );
}

Void TComPrediction::derive1DHash(UChar* pIdx, UInt uiWidth, UInt uiHeight, Pel p1DHashDict[3][64*64])
{
  UChar  crcTable[256] = {
    0x00, 0x3e, 0x7c, 0x42, 0xf8, 0xc6, 0x84, 0xba, 0x95, 0xab, 0xe9, 0xd7, 0x6d, 0x53, 0x11, 0x2f, 0x4f, 0x71, 0x33, 0x0d, 0xb7, 0x89, 0xcb, 0xf5, 0xda, 0xe4, 0xa6, 0x98, 0x22, 0x1c, 0x5e, 0x60,
    0x9e, 0xa0, 0xe2, 0xdc, 0x66, 0x58, 0x1a, 0x24, 0x0b, 0x35, 0x77, 0x49, 0xf3, 0xcd, 0x8f, 0xb1, 0xd1, 0xef, 0xad, 0x93, 0x29, 0x17, 0x55, 0x6b, 0x44, 0x7a, 0x38, 0x06, 0xbc, 0x82, 0xc0, 0xfe,
    0x59, 0x67, 0x25, 0x1b, 0xa1, 0x9f, 0xdd, 0xe3, 0xcc, 0xf2, 0xb0, 0x8e, 0x34, 0x0a, 0x48, 0x76, 0x16, 0x28, 0x6a, 0x54, 0xee, 0xd0, 0x92, 0xac, 0x83, 0xbd, 0xff, 0xc1, 0x7b, 0x45, 0x07, 0x39,
    0xc7, 0xf9, 0xbb, 0x85, 0x3f, 0x01, 0x43, 0x7d, 0x52, 0x6c, 0x2e, 0x10, 0xaa, 0x94, 0xd6, 0xe8, 0x88, 0xb6, 0xf4, 0xca, 0x70, 0x4e, 0x0c, 0x32, 0x1d, 0x23, 0x61, 0x5f, 0xe5, 0xdb, 0x99, 0xa7,
    0xb2, 0x8c, 0xce, 0xf0, 0x4a, 0x74, 0x36, 0x08, 0x27, 0x19, 0x5b, 0x65, 0xdf, 0xe1, 0xa3, 0x9d, 0xfd, 0xc3, 0x81, 0xbf, 0x05, 0x3b, 0x79, 0x47, 0x68, 0x56, 0x14, 0x2a, 0x90, 0xae, 0xec, 0xd2,
    0x2c, 0x12, 0x50, 0x6e, 0xd4, 0xea, 0xa8, 0x96, 0xb9, 0x87, 0xc5, 0xfb, 0x41, 0x7f, 0x3d, 0x03, 0x63, 0x5d, 0x1f, 0x21, 0x9b, 0xa5, 0xe7, 0xd9, 0xf6, 0xc8, 0x8a, 0xb4, 0x0e, 0x30, 0x72, 0x4c,
    0xeb, 0xd5, 0x97, 0xa9, 0x13, 0x2d, 0x6f, 0x51, 0x7e, 0x40, 0x02, 0x3c, 0x86, 0xb8, 0xfa, 0xc4, 0xa4, 0x9a, 0xd8, 0xe6, 0x5c, 0x62, 0x20, 0x1e, 0x31, 0x0f, 0x4d, 0x73, 0xc9, 0xf7, 0xb5, 0x8b,
    0x75, 0x4b, 0x09, 0x37, 0x8d, 0xb3, 0xf1, 0xcf, 0xe0, 0xde, 0x9c, 0xa2, 0x18, 0x26, 0x64, 0x5a, 0x3a, 0x04, 0x46, 0x78, 0xc2, 0xfc, 0xbe, 0x80, 0xaf, 0x91, 0xd3, 0xed, 0x57, 0x69, 0x2b, 0x15
  };
  
  //8-bit CRC with polynomial x^8+x^6+x^3+x^2+1, 0x14D.
  //Chosen based on Koopman, et al. (0xA6 in his notation = 0x14D >> 1):
  //http://www.ece.cmu.edu/~koopman/roses/dsn04/koopman04_crc_poly_embedded.pdf
  //#define POLYNOMIAL 0xB2  // reverse of 0x14D
  //for (int n = 0; n < 256; ++n)
  //{
  //  UChar rem = n;
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  rem = (rem & 1) ? (rem >> 1) ^ POLYNOMIAL : (rem >> 1);
  //  crcTable[n] = rem;
  //}
  
  // 1d hash 2,4,8 
  UInt uiTotal = uiWidth*uiHeight;
  
  Pel pHashDictLast[3][256];
  memset(pHashDictLast, -1, sizeof(Pel)*3*256);
  
  for( UInt uiIdx = 0; uiIdx < uiTotal-8; ++uiIdx)
  {
    UChar rem = 0xff;
    
    rem = crcTable[rem ^ pIdx[uiIdx+0]];
    rem = crcTable[rem ^ pIdx[uiIdx+1]];
    p1DHashDict[0][uiIdx] = pHashDictLast[0][rem];
    pHashDictLast[0][rem] = uiIdx;
    
    rem = crcTable[rem ^ pIdx[uiIdx+2]];
    rem = crcTable[rem ^ pIdx[uiIdx+3]];
    p1DHashDict[1][uiIdx] = pHashDictLast[1][rem];
    pHashDictLast[1][rem] = uiIdx;
    
    rem = crcTable[rem ^ pIdx[uiIdx+4]];
    rem = crcTable[rem ^ pIdx[uiIdx+5]];
    rem = crcTable[rem ^ pIdx[uiIdx+6]];
    rem = crcTable[rem ^ pIdx[uiIdx+7]];
    p1DHashDict[2][uiIdx] = pHashDictLast[2][rem];
    pHashDictLast[2][rem] = uiIdx;
  }
  
  UChar pIdx2[16];
  memcpy( pIdx2, pIdx+uiTotal-8, 8*sizeof(UChar) );
  memset( pIdx2+8, 0, 8*sizeof(UChar) );
  
  for( UInt uiIdx2 = 0; uiIdx2 < 8; ++uiIdx2)
  {
    UInt uiIdx = uiIdx2+uiTotal-8;
    UChar rem = 0xff;
    
    rem = crcTable[rem ^ pIdx2[uiIdx2+0]];
    rem = crcTable[rem ^ pIdx2[uiIdx2+1]];
    p1DHashDict[0][uiIdx] = pHashDictLast[0][rem];
    pHashDictLast[0][rem] = uiIdx;
    
    rem = crcTable[rem ^ pIdx2[uiIdx2+2]];
    rem = crcTable[rem ^ pIdx2[uiIdx2+3]];
    p1DHashDict[1][uiIdx] = pHashDictLast[1][rem];
    pHashDictLast[1][rem] = uiIdx;
    
    rem = crcTable[rem ^ pIdx2[uiIdx2+4]];
    rem = crcTable[rem ^ pIdx2[uiIdx2+5]];
    rem = crcTable[rem ^ pIdx2[uiIdx2+6]];
    rem = crcTable[rem ^ pIdx2[uiIdx2+7]];
    p1DHashDict[2][uiIdx] = pHashDictLast[2][rem];
    pHashDictLast[2][rem] = uiIdx;
  }
}

Void TComPrediction::derive2DHash(UChar* pIdx, UInt uiWidth, UInt uiHeight, UChar* p2DHash, UInt* p2DHashDict, UInt* p2DHashDictLast)
{
  UChar  crcTable[256] = {
    0x00, 0x3e, 0x7c, 0x42, 0xf8, 0xc6, 0x84, 0xba, 0x95, 0xab, 0xe9, 0xd7, 0x6d, 0x53, 0x11, 0x2f, 0x4f, 0x71, 0x33, 0x0d, 0xb7, 0x89, 0xcb, 0xf5, 0xda, 0xe4, 0xa6, 0x98, 0x22, 0x1c, 0x5e, 0x60,
    0x9e, 0xa0, 0xe2, 0xdc, 0x66, 0x58, 0x1a, 0x24, 0x0b, 0x35, 0x77, 0x49, 0xf3, 0xcd, 0x8f, 0xb1, 0xd1, 0xef, 0xad, 0x93, 0x29, 0x17, 0x55, 0x6b, 0x44, 0x7a, 0x38, 0x06, 0xbc, 0x82, 0xc0, 0xfe,
    0x59, 0x67, 0x25, 0x1b, 0xa1, 0x9f, 0xdd, 0xe3, 0xcc, 0xf2, 0xb0, 0x8e, 0x34, 0x0a, 0x48, 0x76, 0x16, 0x28, 0x6a, 0x54, 0xee, 0xd0, 0x92, 0xac, 0x83, 0xbd, 0xff, 0xc1, 0x7b, 0x45, 0x07, 0x39,
    0xc7, 0xf9, 0xbb, 0x85, 0x3f, 0x01, 0x43, 0x7d, 0x52, 0x6c, 0x2e, 0x10, 0xaa, 0x94, 0xd6, 0xe8, 0x88, 0xb6, 0xf4, 0xca, 0x70, 0x4e, 0x0c, 0x32, 0x1d, 0x23, 0x61, 0x5f, 0xe5, 0xdb, 0x99, 0xa7,
    0xb2, 0x8c, 0xce, 0xf0, 0x4a, 0x74, 0x36, 0x08, 0x27, 0x19, 0x5b, 0x65, 0xdf, 0xe1, 0xa3, 0x9d, 0xfd, 0xc3, 0x81, 0xbf, 0x05, 0x3b, 0x79, 0x47, 0x68, 0x56, 0x14, 0x2a, 0x90, 0xae, 0xec, 0xd2,
    0x2c, 0x12, 0x50, 0x6e, 0xd4, 0xea, 0xa8, 0x96, 0xb9, 0x87, 0xc5, 0xfb, 0x41, 0x7f, 0x3d, 0x03, 0x63, 0x5d, 0x1f, 0x21, 0x9b, 0xa5, 0xe7, 0xd9, 0xf6, 0xc8, 0x8a, 0xb4, 0x0e, 0x30, 0x72, 0x4c,
    0xeb, 0xd5, 0x97, 0xa9, 0x13, 0x2d, 0x6f, 0x51, 0x7e, 0x40, 0x02, 0x3c, 0x86, 0xb8, 0xfa, 0xc4, 0xa4, 0x9a, 0xd8, 0xe6, 0x5c, 0x62, 0x20, 0x1e, 0x31, 0x0f, 0x4d, 0x73, 0xc9, 0xf7, 0xb5, 0x8b,
    0x75, 0x4b, 0x09, 0x37, 0x8d, 0xb3, 0xf1, 0xcf, 0xe0, 0xde, 0x9c, 0xa2, 0x18, 0x26, 0x64, 0x5a, 0x3a, 0x04, 0x46, 0x78, 0xc2, 0xfc, 0xbe, 0x80, 0xaf, 0x91, 0xd3, 0xed, 0x57, 0x69, 0x2b, 0x15
  };
  
  memset(p2DHashDictLast, -1, sizeof(UInt)*256);
  
  // 1d hash horz 4
  for( UInt uiY = 0; uiY < uiHeight; ++uiY)
  {
    for( UInt uiX = 0; uiX < (Int)uiWidth - MIN_D2D_WIDTH; ++uiX)
    {
      UChar rem = 0xff;
      UInt uiIdx = uiX + uiY * uiWidth;
      
      rem = crcTable[rem ^ pIdx[uiIdx+0]];
      rem = crcTable[rem ^ pIdx[uiIdx+1]];
      rem = crcTable[rem ^ pIdx[uiIdx+2]];
      rem = crcTable[rem ^ pIdx[uiIdx+3]];
      p2DHash[uiIdx] = rem;
    }
  }
  
  // 1d hash vert 4
  for( UInt uiY = 0; uiY < (Int)uiHeight - MIN_D2D_HEIGHT; ++uiY)
  {
    Bool bTraverse = ( uiY & 1 );

    for( UInt uiX = 0; uiX < (Int)uiWidth - MIN_D2D_WIDTH; ++uiX)
    {
      UChar rem = 0xff;
      Int traX = (bTraverse) ? (uiWidth - MIN_D2D_WIDTH - 1) - uiX : uiX;
      UInt uiIdx = traX + uiY * uiWidth;

      rem = crcTable[rem ^ p2DHash[uiIdx+0]];
      rem = crcTable[rem ^ p2DHash[uiIdx+uiWidth]];
      rem = crcTable[rem ^ p2DHash[uiIdx+uiWidth*2]];
      rem = crcTable[rem ^ p2DHash[uiIdx+uiWidth*3]];
      p2DHashDict[uiIdx] = p2DHashDictLast[rem];
      p2DHashDictLast[rem] = uiIdx;
      p2DHash[uiIdx] = rem;
    }
  }
}
#endif // IDX_1D_SEARCH || DYNAMIC_2D_SEARCH

Void TComPrediction::preCalcPLTIndex(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt uiPLTSize)
{
  Bool bLossless = pcCU->getCUTransquantBypass(0);
  Int iErrorLimit = bLossless ? 0 : 3 * getPLTErrLimit();
  UInt uiPos;

  UInt uiBestIdx = 0;
  UChar useEscapeFlag=0;

  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      UInt uiPLTIdx = 0;
      UInt uiMinError = MAX_UINT;
      while (uiPLTIdx < uiPLTSize)
      {
        UInt uiAbsError = abs(Palette[0][uiPLTIdx] - pSrc[0][uiPos]) + abs(Palette[1][uiPLTIdx] - pSrc[1][uiPos]) + abs(Palette[2][uiPLTIdx] - pSrc[2][uiPos]);
        if (uiAbsError < uiMinError)
        {
          uiBestIdx = uiPLTIdx;
          uiMinError = uiAbsError;
          if (uiMinError == 0)
          {
            break;
          }
        }
        uiPLTIdx++;
      }
      m_cIndexBlock[uiPos] = uiBestIdx;
      if (uiMinError > iErrorLimit)
      {
#if IDX_1D_SEARCH || DYNAMIC_2D_SEARCH
        m_cIndexBlock[uiPos] = uiPLTSize;
#else
        m_cIndexBlock[uiPos] -= MAX_PLT_SIZE;
#endif
        useEscapeFlag=1;
      }
    }
  }

  pcCU->setPLTEscapeSubParts(0, useEscapeFlag,0, pcCU->getDepth(0));
}

Void  TComPrediction::reorderPLT(TComDataCU* pcCU, Pel *pPalette[3], UInt uiNumComp)
{

  UInt uiPLTSizePrev, uiDictMaxSize;
  UInt uiPLTUsedSizePrev;
  Pel * pPalettePrev[3];
  Pel pPaletteTemp[3][MAX_PLT_SIZE];
  ComponentID compBegin = ComponentID(uiNumComp == 2 ? 1 : 0);

  for (UInt comp = compBegin; comp < compBegin + uiNumComp; comp++)
  {
    pPalettePrev[comp] = pcCU->getPLTPred(pcCU, pcCU->getZorderIdxInCU(), comp, uiPLTSizePrev, uiPLTUsedSizePrev);
    for (UInt i = 0; i < MAX_PLT_SIZE; i++)
    {
      pPaletteTemp[comp][i] = pPalette[comp][i];
    }
  }

  pcCU->getPLTPred(pcCU, pcCU->getZorderIdxInCU(), compBegin, uiPLTSizePrev, uiPLTUsedSizePrev);
  uiDictMaxSize = pcCU->getPLTSize(compBegin, 0);

  UInt uiIdxPrev = 0, uiIdxCurr = 0;
  Bool bReused = false;
  Bool bPredicted[MAX_PLT_SIZE + 1], bReusedPrev[MAX_PLT_PRED_SIZE + 1];
  memset(bReusedPrev, 0, sizeof(bReusedPrev));
  memset(bPredicted, 0, sizeof(bPredicted));

  UInt uiNumPLTRceived = uiDictMaxSize, uiNumPLTPredicted = 0;
  for (uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev++)
  {
    bReused = false;
    Int iCounter = 0;
    for (uiIdxCurr = 0; uiIdxCurr < uiDictMaxSize; uiIdxCurr++)
    {
      iCounter = 0;

      for (UInt comp = compBegin; comp < compBegin + uiNumComp; comp++)
      {
        if (pPalettePrev[comp][uiIdxPrev] == pPalette[comp][uiIdxCurr])
        {
          iCounter++;
        }
      }
      if (iCounter == uiNumComp)
      {
        bReused = true;
        break;
      }
    }
    bReusedPrev[uiIdxPrev] = bReused;
    bPredicted[uiIdxCurr] = bReused;
    if (bPredicted[uiIdxCurr])
    {
      uiNumPLTRceived--;
      uiNumPLTPredicted++;
    }
  }

  for (uiIdxPrev = 0; uiIdxPrev < MAX_PLT_PRED_SIZE; uiIdxPrev++)
  {
    for (UInt comp = compBegin; comp < compBegin + uiNumComp; comp++)
    {
      pcCU->setPrevPLTReusedFlagSubParts(comp, bReusedPrev[uiIdxPrev], uiIdxPrev, 0, pcCU->getDepth(0));
    }
  }
  uiIdxCurr = 0;
  for (UInt uiPrevIdx = 0; uiPrevIdx < uiPLTSizePrev; uiPrevIdx++)
  {
    if (bReusedPrev[uiPrevIdx])
    {
      for (UInt comp = compBegin; comp < compBegin + uiNumComp; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPalettePrev[comp][uiPrevIdx];
      }
      uiIdxCurr++;
    }
  }

  for (UInt uiIdx = 0; uiIdx < uiDictMaxSize; uiIdx++)
  {
    if (bPredicted[uiIdx] == 0)
    {
      for (UInt comp = compBegin; comp < compBegin + uiNumComp; comp++)
      {
        pPalette[comp][uiIdxCurr] = pPaletteTemp[comp][uiIdx];
      }
      uiIdxCurr++;
    }
  }
}

Void  TComPrediction::derivePLTLossy( TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3],  UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize, TComRdCost *pcCost )
{
  Int iErrorLimit = getPLTErrLimit();
  UInt uiTotalSize = uiHeight*uiWidth;
  SortingElement *psList = new SortingElement [uiTotalSize];
  SortingElement sElement;
  UInt uiDictMaxSize = MAX_PLT_SIZE;
  SortingElement *pListSort = new SortingElement [MAX_PLT_SIZE + 1];
  UInt uiIdx = 0;
  UInt uiPos;
  Int last = -1;

  for( UInt uiY = 0; uiY < uiHeight; uiY++ )
  {
    for( UInt uiX = 0; uiX < uiWidth; uiX++ )
    {
      uiPos = uiY * uiWidth+ uiX;
      sElement.setAll (pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);

      Int besti = last, bestSAD = (last == -1) ? MAX_UINT : psList[last].getSAD(sElement);
      if( bestSAD )
      {
        for( Int i = uiIdx-1; i>=0; i-- )
        {
          UInt sad = psList[i].getSAD(sElement);
          if( sad < bestSAD )
          {
            bestSAD = sad;
            besti = i;
            if( !sad ) break;
          }
        }
      }

      if (besti>=0 && psList[besti].almostEqualData(sElement, iErrorLimit))
      {
        psList[besti].addElement(sElement);
        last = besti;
      }
      else
      {
        psList[uiIdx].copyDataFrom(sElement);
        psList[uiIdx].uiCnt = 1;
        last = uiIdx;
        uiIdx ++;
      }
    }
  }

  for (Int i = 0; i < uiDictMaxSize; i++)
  {
    pListSort[i].uiCnt  = 0;
    pListSort[i].setAll(0, 0, 0) ;
  }

  //bubble sorting
  uiDictMaxSize = 1;
  for (Int i = 0; i < uiIdx; i++)
  {
    if( psList[i].uiCnt > pListSort[uiDictMaxSize-1].uiCnt )
    {
      Int j;
      for (j = uiDictMaxSize; j > 0; j--)
      {
        if (psList[i].uiCnt > pListSort[j-1].uiCnt)
        {
          pListSort[j].copyAllFrom (pListSort[j-1]);
          uiDictMaxSize = std::min(uiDictMaxSize+1,(UInt)MAX_PLT_SIZE);
        }
        else
        {
          break;
        }
      }
      pListSort[j].copyAllFrom (psList[i]) ;
    }
  }

  uiPLTSize = 0;
  Pel *pPred[3]  = { pcCU->getLastPLTInLcuFinal(0), pcCU->getLastPLTInLcuFinal(1), pcCU->getLastPLTInLcuFinal(2) };
  Double bitCost = pcCost->getLambda() * 24;

  for (Int i = 0; i < MAX_PLT_SIZE; i++)
  {
    if( pListSort[i].uiCnt )
    {
      Int iHalf = pListSort[i].uiCnt>>1;
      Palette[0][uiPLTSize] = (pListSort[i].uiSumData[0]+iHalf)/pListSort[i].uiCnt;
      Palette[1][uiPLTSize] = (pListSort[i].uiSumData[1]+iHalf)/pListSort[i].uiCnt;
      Palette[2][uiPLTSize] = (pListSort[i].uiSumData[2]+iHalf)/pListSort[i].uiCnt;

      Int best = -1;
      if( iErrorLimit )
      {
        Double pal[3] = { pListSort[i].uiSumData[0]/(Double)pListSort[i].uiCnt,
                          pListSort[i].uiSumData[1]/(Double)pListSort[i].uiCnt,
                          pListSort[i].uiSumData[2]/(Double)pListSort[i].uiCnt };

        Double err      = pal[0] - Palette[0][uiPLTSize];
        Double bestCost = err*err;
        err = pal[1] - Palette[1][uiPLTSize]; bestCost += err*err;
        err = pal[2] - Palette[2][uiPLTSize]; bestCost += err*err;
        bestCost = bestCost * pListSort[i].uiCnt + bitCost;

        for(int t=0; t<pcCU->getLastPLTInLcuSizeFinal(0); t++)
        {
          err = pal[0] - pPred[0][t];
          Double cost = err*err;
          err = pal[1] - pPred[1][t]; cost += err*err;
          err = pal[2] - pPred[2][t]; cost += err*err;
          cost *= pListSort[i].uiCnt;
          if(cost < bestCost)
          {
            best = t;
            bestCost = cost;
          }
        }
        if( best != -1 )
        {
          Palette[0][uiPLTSize] = pPred[0][best];
          Palette[1][uiPLTSize] = pPred[1][best];
          Palette[2][uiPLTSize] = pPred[2][best];
        }
      }

      Bool bDuplicate = false;
      if( pListSort[i].uiCnt == 1 && best == -1 )
      {
        bDuplicate = true;
      }
      else
      {
        for( Int t=0; t<uiPLTSize; t++)
        {
          if( Palette[0][uiPLTSize] == Palette[0][t] && Palette[1][uiPLTSize] == Palette[1][t] && Palette[2][uiPLTSize] == Palette[2][t] )
          {
            bDuplicate = true;
            break;
          }
        }
      }
      if( !bDuplicate ) uiPLTSize++;
    }
    else
    {
      break;
    }
  }

  delete [] psList;
  delete [] pListSort;
}

Void TComPrediction::derivePLTLossless(TComDataCU* pcCU, Pel *Palette[3], Pel* pSrc[3], UInt uiWidth, UInt uiHeight, UInt uiStride, UInt &uiPLTSize)
{
  std::vector<SortingElement> psList;
  SortingElement sElement;
  UInt uiIdx = 0;
  UInt uiPos;
  for (UInt uiY = 0; uiY < uiHeight; uiY++)
  {
    for (UInt uiX = 0; uiX < uiWidth; uiX++)
    {
      uiPos = uiY * uiWidth + uiX;
      sElement.setAll(pSrc[0][uiPos], pSrc[1][uiPos], pSrc[2][uiPos]);
      Int i = 0;
      for (i = uiIdx - 1; i >= 0; i--)
      {
        if( psList[i].uiData[0] == sElement.uiData[0] && psList[i].uiData[1] == sElement.uiData[1] && psList[i].uiData[2] == sElement.uiData[2] )
        {
          psList[i].uiCnt++;
          break;
        }
      }
      if (i == -1)
      {
        psList.push_back(sElement);
        psList[uiIdx].uiCnt++;
        uiIdx++;
      }
    }
  }
  //insertion sort, high frequency -> low frequency
  std::stable_sort(psList.begin(), psList.end());
  UInt uiPLTSizePrev;
  UInt uiPLTUsedSizePrev;
  Pel *pPalettePrev[3];
  for (UInt comp = 0; comp < 3; comp++)
  {
    pPalettePrev[comp] = pcCU->getPLTPred(pcCU, pcCU->getZorderIdxInCU(), comp, uiPLTSizePrev, uiPLTUsedSizePrev);
  }

  uiPLTSize = 0;
  for (Int i = 0; i < uiIdx; i++)
  {
    for (Int j = i - 1; j >= 0; j--)
    {
      if (psList[j].uiCnt && psList[i].almostEqualData(psList[j], 0))
      {
        psList[i].uiCnt = 0;
        break;
      }
    }

    Bool includeIntoPLT = true;
    if( psList[i].uiCnt == 1 )
    {
      includeIntoPLT = false;
      for( UInt uiIdxPrev = 0; uiIdxPrev < uiPLTSizePrev; uiIdxPrev++ )
      {
        UInt iCounter = 0;

        for( UInt comp = 0; comp < 3; comp++ )
        {
          if( psList[i].uiData[comp] == pPalettePrev[comp][uiIdxPrev] )
          {
            iCounter++;
          }
          else
          {
            break;
          }
        }
        if( iCounter == 3 )
        {
          includeIntoPLT = true;
          break;
        }
      }
    }

    if( includeIntoPLT && psList[i].uiCnt)
    {
      Palette[0][uiPLTSize] = psList[i].uiData[0];
      Palette[1][uiPLTSize] = psList[i].uiData[1];
      Palette[2][uiPLTSize] = psList[i].uiData[2];
      uiPLTSize++;
      if (uiPLTSize == MAX_PLT_SIZE)
      {
        break;
      }
    }
  }
}

Void TComPrediction::calcPixelPred(TComDataCU* pcCU, Pel* pOrg [3], Pel *pPalette[3], Pel* pValue, Pel*paPixelValue[3], Pel * paRecoValue[3], UInt uiWidth, UInt uiHeight,  UInt uiStrideOrg, UInt uiIdx )
{
  Bool bLossless = pcCU->getCUTransquantBypass (0);
  Int iQP[3];
  Int iQPrem[3];
  Int iQPper[3];
  Int quantiserScale[3];
  Int quantiserRightShift[3];
  Int rightShiftOffset[3];
  Int InvquantiserRightShift[3];
  Int iAdd[3];
  for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch++)
  {
    QpParam cQP(*pcCU, ComponentID(ch));
    iQP[ch] = cQP.Qp;
    iQPrem[ch] = iQP[ch] % 6;
    iQPper[ch] = iQP[ch] / 6;
    quantiserScale[ch] = g_quantScales[iQPrem[ch]];
    quantiserRightShift[ch] = QUANT_SHIFT + iQPper[ch];
    rightShiftOffset[ch] = 1 << (quantiserRightShift[ch] - 1);
    InvquantiserRightShift[ch] = (IQUANT_SHIFT - iQPper[ch]);
    iAdd[ch] = InvquantiserRightShift[ch] == 0 ? 0 : 1 << (InvquantiserRightShift[ch] - 1);
  }

  UInt uiY, uiX;
  uiY = uiIdx / uiWidth;
  uiX = uiIdx % uiWidth;
  UInt uiScanIdx = uiY * uiStrideOrg + uiX;
  UInt uiYIdxRaster = pcCU->getPLTScanRotationModeFlag(0)? (uiX * uiStrideOrg + uiY) : (uiY * uiStrideOrg + uiX);
  UInt uiMaxBit[3];
  pcCU->xCalcMaxBits(pcCU, uiMaxBit);
  if (bLossless)
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
      paPixelValue[ch][uiScanIdx] =  pOrg[ch][uiYIdxRaster];
      paRecoValue[ch][uiYIdxRaster] = pOrg[ch][uiYIdxRaster];
    }
  }
  else
  {
    for (UInt ch = 0; ch < MAX_NUM_COMPONENT; ch ++)
    {
      paPixelValue[ch][uiScanIdx] = Pel(ClipBD<Int>((pOrg[ch][uiYIdxRaster] * quantiserScale[ch] + rightShiftOffset[ch]) >> quantiserRightShift[ch], uiMaxBit[ch]));
      paRecoValue[ch][uiYIdxRaster] = (paPixelValue[ch] [uiScanIdx]*g_invQuantScales[iQPrem[ch]] + iAdd[ch])>>InvquantiserRightShift[ch];
      paRecoValue[ch][uiYIdxRaster] = Pel(ClipBD<Int>(paRecoValue[ch][uiYIdxRaster], g_bitDepth[ch? 1:0]));
      //if( pcCU->getAddr() == 171 && pcCU->getCUPelX() == 728 && pcCU->getCUPelY() == 544 )
      //{
      //  fprintf(stderr, "idx=%d, %d, %3d, %3d, %d, %d, %d, %d\n", uiIdx, ch, pOrg[ch][uiYIdxRaster], paRecoValue[ch][uiYIdxRaster], quantiserScale[ch], rightShiftOffset[ch], quantiserRightShift[ch], uiMaxBit[ch]);
      //}
    }
  }
}

Bool TComPrediction::calLeftRun(Pel* pValue, UChar* pSPoint, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  UInt uiIdx = uiStartPos;
  Pel *pcIndexBlock = m_cIndexBlock;
  while (uiIdx < uiTotal)
  {
    UInt uiTraIdx = m_puiScanOrder[uiIdx];  //unified position variable (raster scan)
    pValue[uiTraIdx] = pcIndexBlock[uiTraIdx] < 0 ? pcIndexBlock[uiTraIdx] + MAX_PLT_SIZE : pcIndexBlock[uiTraIdx];
    Bool bMismatch = (pcIndexBlock[uiTraIdx] < 0);
    if (bMismatch && uiIdx == uiStartPos)
    {
      return false;
    }
    UInt leftTraIdx = uiIdx ? m_puiScanOrder[uiIdx - 1] : 0;
    if (uiIdx > uiStartPos && (PLT_ESCAPE != pSPoint[leftTraIdx]) && pValue[uiTraIdx] == pValue[leftTraIdx] && !bMismatch)
    {
      uiRun++;
    }
    else if (uiIdx > uiStartPos)
    {
      break;
    }
    uiIdx++;
  }
  return true;
}

Bool  TComPrediction::calAboveRun(Pel* pValue, UChar* pSPoint, UInt uiWidth, UInt uiStartPos, UInt uiTotal, UInt &uiRun)
{
  UInt uiIdx = uiStartPos;
  UInt uiY = 0;
  Bool valid = false;
  Pel *pcIndexBlock = m_cIndexBlock;
  UInt uiTraIdx = m_puiScanOrder[uiIdx];  //unified position variable (raster scan)

  while (uiIdx < uiTotal)
  {
    uiY = uiTraIdx / uiWidth;
    if (uiY == 0)
    {
      return false;
    }
    UInt uiStride = uiWidth;
    uiTraIdx = m_puiScanOrder[uiIdx];  //unified position variable (raster scan)

    pValue[uiTraIdx] = pcIndexBlock[uiTraIdx] < 0 ? pcIndexBlock[uiTraIdx] + MAX_PLT_SIZE : pcIndexBlock[uiTraIdx];
    Bool bMismatch = (pcIndexBlock[uiTraIdx] < 0);
    if (uiRun == 0 && bMismatch)
    {
      return false;
    }
    if ((PLT_ESCAPE != pSPoint[uiTraIdx - uiStride]) && pValue[uiTraIdx] == pValue[uiTraIdx - uiStride] && !bMismatch)
    {
      uiRun++;
      valid = true;
    }
    else
    {
      break;
    }
    uiIdx++;
  }
  return valid;
}

Void  TComPrediction::rotationScan( Pel* pLevel, UInt uiWidth, UInt uiHeight, Bool isInverse )
{
  Pel tmpLevel;
  UInt uiPos = 0;
  UInt* puiScanOrder = g_scanOrder[SCAN_UNGROUPED][SCAN_VER][g_aucConvertToBit[uiWidth] + 2][g_aucConvertToBit[uiHeight] + 2];

  for (UInt j = 1; j < uiHeight; j++)
  {
    uiPos += j;
    for (UInt i = j; i < uiWidth; i++)
    {
      tmpLevel = pLevel[uiPos];
      pLevel[uiPos] = pLevel[puiScanOrder[uiPos]];
      pLevel[puiScanOrder[uiPos]] = tmpLevel;
      uiPos++;
    }
  }
}
#endif
//! \}
